h1.logo, div.section#wdom > h1 {
  font-family: "Raleway";
  font-weight: 500;
}

a.headerlink {
  color: rgba(0, 0, 0, 0.1);
}

div.sphinxsidebarwrapper p.blurb {
  font-family: Lato, sans-serif;
}

div.sphinxsidebar li.toctree-l1 {
  font-family: Lato, sans-serif;
}

body {
  background-color: #fafafa
}

.search-btn {
  padding: 0 1em;
  font-family: Lato, sans-serif;
  font-weight: normal;
  line-height: normal;
  align-self: stretch;
}
